﻿Namespace LogisticWithMulti
    Friend Class LogisticMultiProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin Logistic Regression Classification with Multi-Swarm demo")

            Dim numFeatures As Integer = 5
            Dim numRows As Integer = 10000
            Dim seed As Integer = 0

            Console.WriteLine(vbCrLf & "Generating " & numRows & " artificial data items with " & numFeatures & " features")
            Dim allData As Double()() = MakeAllData(numFeatures, numRows, seed)
            Console.WriteLine("Done")

            Console.WriteLine(vbCrLf & "Creating train (80%) and test (20%) matrices")
            Dim trainData As Double()() = Nothing
            Dim testData As Double()() = Nothing
            MakeTrainTest(allData, 0.8, seed, trainData, testData)
            Console.WriteLine("Done")

            Console.WriteLine(vbCrLf & "Training data: " & vbCrLf)
            ShowData(trainData, 4, 2, True)

            Console.WriteLine(vbCrLf & "Test data: " & vbCrLf)
            ShowData(testData, 3, 2, True)

            Console.WriteLine("Creating Logistic Regression classifier" & vbCrLf)
            Dim lc As New LogisticClassifier(numFeatures)

            Dim numSwarms As Integer = 4
            Dim numParticles As Integer = 3
            Dim maxEpochs As Integer = 100
            Console.WriteLine("Setting numSwarms = " & numSwarms)
            Console.WriteLine("Setting numParticles = " & numParticles)
            Console.WriteLine("Setting maxEpochs = " & maxEpochs)
            Console.WriteLine(vbCrLf & "Starting training")
            Dim bestWeights() As Double = lc.Train(trainData, maxEpochs, numSwarms, numParticles)
            Console.WriteLine("Training complete" & vbCrLf)
            Console.WriteLine("Best weights found:")
            ShowVector(bestWeights, 4, True)

            Dim trainAcc As Double = lc.Accuracy(trainData, bestWeights)
            Console.WriteLine("Accuracy on training data = " & trainAcc.ToString("F4"))

            Dim testAcc As Double = lc.Accuracy(testData, bestWeights)
            Console.WriteLine("Accuracy on test data = " & testAcc.ToString("F4"))

            Console.WriteLine(vbCrLf & "End Logistic Classification with Multi-Swarm demo" & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Function MakeAllData(ByVal numFeatures As Integer, ByVal numRows As Integer, ByVal seed As Integer) As Double()()
            Dim rnd As New Random(seed)
            Dim weights(numFeatures) As Double ' inc. b0
            For i = 0 To weights.Length - 1
                weights(i) = 20.0 * rnd.NextDouble - 10.0 ' -10.0 to +10.0
            Next i

            Dim result(numRows - 1)() As Double ' allocate matrix
            For i = 0 To numRows - 1
                result(i) = New Double(numFeatures) {} ' Y in last column
            Next i

            For i = 0 To numRows - 1 ' for each row
                Dim y As Double = weights(0) ' the b0
                For j = 0 To numFeatures - 1 ' each feature / column except last
                    Dim x As Double = 20.0 * rnd.NextDouble - 10.0 ' random X in 10.0, +10.0
                    result(i)(j) = x ' store x
                    Dim wx As Double = x * weights(j + 1) ' weight * x
                    y += wx ' accumulate to get Y
                Next j

                If y > 2.0 Then
                    result(i)(numFeatures) = 1.0 ' store y in last column
                Else
                    result(i)(numFeatures) = 0.0
                End If
            Next i
            Console.WriteLine("Data generation weights:")
            ShowVector(weights, 4, False)

            Return result
        End Function ' MakeAllData

        Private Shared Sub MakeTrainTest(ByVal allData()() As Double, ByVal trainPct As Double, ByVal seed As Integer, <System.Runtime.InteropServices.Out> ByRef trainData()() As Double, <System.Runtime.InteropServices.Out> ByRef testData()() As Double)
            Dim rnd As New Random(seed)
            Dim totRows As Integer = allData.Length
            Dim numTrainRows As Integer = CInt(Fix(totRows * trainPct)) ' usually 0.80
            Dim numTestRows As Integer = totRows - numTrainRows
            trainData = New Double(numTrainRows - 1)() {}
            testData = New Double(numTestRows - 1)() {}

            Dim copy(allData.Length - 1)() As Double ' ref copy of all data
            For i = 0 To copy.Length - 1
                copy(i) = allData(i)
            Next i

            For i = 0 To copy.Length - 1 ' scramble order
                Dim r As Integer = rnd.Next(i, copy.Length) ' use Fisher-Yates
                Dim tmp() As Double = copy(r)
                copy(r) = copy(i)
                copy(i) = tmp
            Next i
            For i = 0 To numTrainRows - 1
                trainData(i) = copy(i)
            Next i

            For i = 0 To numTestRows - 1
                testData(i) = copy(i + numTrainRows)
            Next i
        End Sub ' MakeTrainTest

        Private Shared Sub ShowData(ByVal data As Double()(), ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
            For i = 0 To numRows - 1
                If indices = True Then
                    Console.Write("[" & i.ToString.PadLeft(2) & "]  ")
                End If
                For j = 0 To data(i).Length - 1
                    Dim v As Double = data(i)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "    ")
                Next j
                Console.WriteLine("")
            Next i
            Console.WriteLine(". . .")
            Dim lastRow As Integer = data.Length - 1
            If indices = True Then
                Console.Write("[" & lastRow.ToString.PadLeft(2) & "]  ")
            End If
            For j = 0 To data(lastRow).Length - 1
                Dim v As Double = data(lastRow)(j)
                If v >= 0.0 Then
                    Console.Write(" ") ' '+'
                End If
                Console.Write(v.ToString("F" & decimals) & "    ")
            Next j
            Console.WriteLine(vbCrLf)
        End Sub

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal newLine As Boolean)
            For i = 0 To vector.Length - 1
                Console.Write(vector(i).ToString("F" & decimals) & " ")
            Next i
            Console.WriteLine("")
            If newLine = True Then
                Console.WriteLine("")
            End If
        End Sub
    End Class ' Program

    Public Class LogisticClassifier
        Private numFeatures As Integer ' number of independent variables aka features
        Private weights() As Double ' weights0 = b0 = is the constant
        Private rnd As Random

        Public Sub New(ByVal numFeatures As Integer)
            Me.numFeatures = numFeatures ' number features/predictors
            Me.weights = New Double(numFeatures) {} ' 0 = b0 constant
            Me.rnd = New Random(0) ' pDeath, pImmigrate, sequencing
        End Sub

        Public Function Train(ByVal trainData As Double()(), ByVal maxEpochs As Integer, ByVal numSwarms As Integer, ByVal numParticles As Integer) As Double()
            Dim [dim] As Integer = numFeatures + 1 ' plus the b0 constant
            Dim minX As Double = -10.0
            Dim maxX As Double = 10.0

            Dim ms As New MultiSwarm(numSwarms, numParticles, [dim]) ' array of arrays of Particles
            ' supply error now
            For i = 0 To numSwarms - 1 ' each swarm
                For j = 0 To numParticles - 1 ' each particle
                    Dim p As Particle = ms.swarms(i).particles(j)
                    p.error = [Error](trainData, p.position) ' add error
                    p.bestError = p.error
                    Array.Copy(p.position, p.bestPosition, [dim])
                    If p.error < ms.swarms(i).bestError Then ' swarm best?
                        ms.swarms(i).bestError = p.error
                        Array.Copy(p.position, ms.swarms(i).bestPosition, [dim])
                    End If
                    If p.error < ms.bestError Then ' global best?
                        ms.bestError = p.error
                        Array.Copy(p.position, ms.bestPosition, [dim])
                    End If
                Next j ' j
            Next i ' i

            Dim epoch As Integer = 0
            Dim w As Double = 0.729 ' inertia
            Dim c1 As Double = 1.49445 ' particle-cogntive
            Dim c2 As Double = 1.49445 ' swarm-social
            Dim c3 As Double = 0.3645 ' multiswarm-global
            Dim pDeath As Double = 1.0 / maxEpochs ' prob of particle death
            Dim pImmigrate As Double = 1.0 / maxEpochs ' prob of particle immigration

            Dim sequence(numParticles - 1) As Integer ' process particles in random order
            For i = 0 To sequence.Length - 1
                sequence(i) = i
            Next i

            Do While epoch < maxEpochs
                epoch += 1

                If epoch Mod 10 = 0 AndAlso epoch < maxEpochs Then ' progress
                    Console.Write("Epoch = " & epoch)
                    Console.WriteLine("   Best error = " & ms.bestError.ToString("F4"))
                End If

                For i = 0 To numSwarms - 1 ' each swarm
                    Shuffle(sequence) ' move particles in random sequence
                    For pj = 0 To numParticles - 1 ' each particle
                        Dim j = sequence(pj)
                        Dim p As Particle = ms.swarms(i).particles(j) ' ref to curr particle

                        ' 1. update velocity
                        For k = 0 To [dim] - 1
                            Dim r1 As Double = rnd.NextDouble
                            Dim r2 As Double = rnd.NextDouble
                            Dim r3 As Double = rnd.NextDouble
                            p.velocity(k) = (w * p.velocity(k)) + (c1 * r1 * (p.bestPosition(k) - p.position(k))) + (c2 * r2 * (ms.swarms(i).bestPosition(k) - p.position(k))) + (c3 * r3 * (ms.bestPosition(k) - p.position(k)))

                            If p.velocity(k) < minX Then
                                p.velocity(k) = minX
                            ElseIf p.velocity(k) > maxX Then
                                p.velocity(k) = maxX
                            End If
                        Next k ' k

                        ' 2. update position
                        For k = 0 To [dim] - 1
                            p.position(k) += p.velocity(k)
                            ' constrain all xi in new position
                            If p.position(k) < minX Then
                                p.position(k) = minX
                            ElseIf p.position(k) > maxX Then
                                p.position(k) = maxX
                            End If
                        Next k

                        ' 3. update error at new position
                        p.error = [Error](trainData, p.position) ' expensive
                        ' update best particle, swarm, and best global error if necessary
                        If p.error < p.bestError Then ' new best position for particle?
                            p.bestError = p.error
                            Array.Copy(p.position, p.bestPosition, [dim])
                        End If
                        If p.error < ms.swarms(i).bestError Then ' new best for swarm?
                            ms.swarms(i).bestError = p.error
                            Array.Copy(p.position, ms.swarms(i).bestPosition, [dim])
                        End If
                        If p.error < ms.bestError Then ' new global best?
                            ms.bestError = p.error
                            Array.Copy(p.position, ms.bestPosition, [dim])
                        End If

                        ' 4. die?
                        Dim p1 As Double = rnd.NextDouble
                        'double pDeath = (maxProbDeath / maxEpochs) * epoch; // pDeath increases over time
                        If p1 < pDeath Then
                            Dim q As New Particle([dim]) ' a replacement
                            q.error = [Error](trainData, q.position)
                            Array.Copy(q.position, q.bestPosition, [dim])
                            q.bestError = q.error

                            If q.error < ms.swarms(i).bestError Then ' new best swarm error by pure luck?
                                ms.swarms(i).bestError = q.error
                                Array.Copy(q.position, ms.swarms(i).bestPosition, [dim])

                                If q.error < ms.bestError Then ' best global error?
                                    ms.bestError = q.error
                                    Array.Copy(q.position, ms.bestPosition, [dim])
                                End If
                            End If

                            ms.swarms(i).particles(j) = q ' die!
                        End If ' die

                        ' 5. immigrate?
                        Dim p2 As Double = rnd.NextDouble
                        If p2 < pImmigrate Then
                            Dim ii As Integer = rnd.Next(0, numSwarms) ' rnd swarm
                            Dim jj As Integer = rnd.Next(0, numParticles) ' rnd particle
                            Dim q As Particle = ms.swarms(ii).particles(jj) ' q points to other

                            ms.swarms(i).particles(j) = q
                            ms.swarms(ii).particles(jj) = p ' the exchange
                            ' p now points to other, q points to curr

                            If q.error < ms.swarms(i).bestError Then ' curr has new position
                                ms.swarms(i).bestError = q.error
                                Array.Copy(q.position, ms.swarms(i).bestPosition, [dim])
                            End If
                            If p.error < ms.swarms(ii).bestError Then ' other has new position
                                ms.swarms(ii).bestError = p.error
                                Array.Copy(p.position, ms.swarms(ii).bestPosition, [dim])
                            End If
                        End If ' immigrate
                    Next pj ' j
                Next i ' i
            Loop ' while

            Return ms.bestPosition
        End Function

        Private Sub Shuffle(ByVal sequence() As Integer)
            For i = 0 To sequence.Length - 1
                Dim r As Integer = rnd.Next(i, sequence.Length)
                Dim tmp As Integer = sequence(r)
                sequence(r) = sequence(i)
                sequence(i) = tmp
            Next i
        End Sub

        Public Function [Error](ByVal trainData As Double()(), ByVal weights() As Double) As Double
            ' mean squared error using supplied weights
            Dim yIndex As Integer = trainData(0).Length - 1 ' y-value (0/1) is last column
            Dim sumSquaredError As Double = 0.0
            For i = 0 To trainData.Length - 1 ' each data
                Dim computed As Double = ComputeOutput(trainData(i), weights)
                Dim desired As Double = trainData(i)(yIndex) ' ex: 0.0 or 1.0
                sumSquaredError += (computed - desired) * (computed - desired)
            Next i
            Return sumSquaredError / trainData.Length
        End Function

        Public Function ComputeOutput(ByVal dataItem() As Double, ByVal weights() As Double) As Double
            Dim z As Double = 0.0
            z += weights(0) ' start w/ the b0 constant
            For i = 0 To weights.Length - 2 ' data might include Y
                z += (weights(i + 1) * dataItem(i)) ' skip first weight
            Next i
            Return 1.0 / (1.0 + Math.Exp(-z))
        End Function

        Public Function ComputeDependent(ByVal dataItem() As Double, ByVal weights() As Double) As Integer
            Dim sum As Double = ComputeOutput(dataItem, weights)

            If sum <= 0.5 Then
                Return 0
            Else
                Return 1
            End If
        End Function

        Public Function Accuracy(ByVal trainData As Double()(), ByVal weights() As Double) As Double
            Dim numCorrect As Integer = 0
            Dim numWrong As Integer = 0
            Dim yIndex As Integer = trainData(0).Length - 1
            For i = 0 To trainData.Length - 1
                Dim computed As Double = ComputeDependent(trainData(i), weights) ' implicit cast
                Dim desired As Double = trainData(i)(yIndex) ' 0.0 or 1.0

                Dim closeness As Double = 0.00000001 ' 'epsilon'
                If Math.Abs(computed - desired) < closeness Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If
            Next i
            Return (numCorrect * 1.0) / (numWrong + numCorrect)
        End Function

        ' ===== nested classes for MSO =====

        Private Class Particle
            Private Shared rnd As New Random(0) ' shared by all instances
            Public position() As Double
            Public velocity() As Double
            Public [error] As Double
            Public bestPosition() As Double
            Public bestError As Double
            Private minX As Double = -10.0
            Private maxX As Double = 10.0

            Public Sub New(ByVal [dim] As Integer)
                position = New Double([dim] - 1) {}
                velocity = New Double([dim] - 1) {}
                bestPosition = New Double([dim] - 1) {}
                For k = 0 To [dim] - 1
                    position(k) = (maxX - minX) * rnd.NextDouble + minX
                    velocity(k) = ((0.1 * maxX) - (0.1 * minX)) * rnd.NextDouble + (0.1 * minX)
                Next k
                [error] = Double.MaxValue ' temp dummy
                Array.Copy(position, bestPosition, [dim]) ' not necessary
                bestError = [error] ' temp
            End Sub
        End Class ' Particle

        Private Class Swarm
            ' an array of particles + swarm best error
            Public particles() As Particle
            Public bestPosition() As Double ' of any particle
            Public bestError As Double

            Public Sub New(ByVal numParticles As Integer, ByVal [dim] As Integer)
                particles = New Particle(numParticles - 1) {}
                For i = 0 To numParticles - 1
                    particles(i) = New Particle([dim])
                Next i

                bestPosition = New Double([dim] - 1) {}
                bestError = Double.MaxValue
            End Sub ' ctor
        End Class ' Swarm

        Private Class MultiSwarm
            ' an array of Swarms + global best error
            Public swarms() As Swarm
            Public bestPosition() As Double ' of any particle in any swarm
            Public bestError As Double

            Public Sub New(ByVal numSwarms As Integer, ByVal numParticles As Integer, ByVal [dim] As Integer)
                swarms = New Swarm(numSwarms - 1) {}
                For i = 0 To numSwarms - 1
                    swarms(i) = New Swarm(numParticles, [dim])
                Next i
                bestPosition = New Double([dim] - 1) {}
                bestError = Double.MaxValue
            End Sub ' ctor
        End Class ' MultiSwarm
    End Class ' LogisticClassifier
End Namespace ' ns
